@@ -35,6 +35,10 @@ THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #define SNA_XVMC 1
 #endif
 
+#define FOURCC_XVMC (('C' << 24) + ('M' << 16) + ('V' << 8) + 'X')
+#define FOURCC_RGB565 ((16 << 24) + ('B' << 16) + ('G' << 8) + 'R')
+#define FOURCC_RGB888 ((24 << 24) + ('B' << 16) + ('G' << 8) + 'R')
+
 /*
  * Below, a dummy picture type that is used in XvPutImage
  * only to do an overlay update.
@@ -49,6 +53,22 @@ THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 	XvTopToBottom \
 }
 
+#define XVMC_RGB565 { \
+	FOURCC_RGB565, XvRGB, LSBFirst, \
+	{'P', 'A', 'S', 'S', 'T', 'H', 'R', 'O', 'U', 'G', 'H', 'R', 'G', 'B', '1', '6'}, \
+	16, XvPacked, 1, 16, 0x1f<<11, 0x3f<<5, 0x1f<<0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
+	{'B', 'G', 'R', 'X', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, \
+	XvTopToBottom \
+}
+
+#define XVMC_RGB888 { \
+	FOURCC_RGB888, XvRGB, LSBFirst, \
+	{'P', 'A', 'S', 'S', 'T', 'H', 'R', 'O', 'U', 'G', 'H', 'R', 'G', 'B', '2', '4'}, \
+	32, XvPacked, 1, 24, 0xff<<16, 0xff<<8, 0xff<<0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
+	{'B', 'G', 'R', 'X', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, \
+	XvTopToBottom \
+}
+
 struct sna_video {
 	struct sna *sna;
 
@@ -78,8 +98,10 @@ struct sna_video {
 	bool textured;
 	Rotation rotation;
 	int plane;
+	struct kgem_bo *bo;
 
 	int SyncToVblank;	/* -1: auto, 0: off, 1: on */
+	int AlwaysOnTop;
 };
 
 struct sna_video_frame {
@@ -109,14 +131,22 @@ void sna_video_textured_setup(struct sna *sna, ScreenPtr screen);
 void sna_video_destroy_window(WindowPtr win);
 
 XvAdaptorPtr sna_xv_adaptor_alloc(struct sna *sna);
+int sna_xv_fixup_formats(ScreenPtr screen,
+			 XvFormatPtr formats,
+			 int num_formats);
 int sna_xv_alloc_port(unsigned long port, XvPortPtr in, XvPortPtr *out);
 int sna_xv_free_port(XvPortPtr port);
 
-#define FOURCC_XVMC     (('C' << 24) + ('M' << 16) + ('V' << 8) + 'X')
-
 static inline int xvmc_passthrough(int id)
 {
-	return id == FOURCC_XVMC;
+	switch (id) {
+	case FOURCC_XVMC:
+	case FOURCC_RGB565:
+	case FOURCC_RGB888:
+		return true;
+	default:
+		return false;
+	}
 }
 
 static inline int is_planar_fourcc(int id)
@@ -126,18 +156,15 @@ static inline int is_planar_fourcc(int id)
 	case FOURCC_I420:
 	case FOURCC_XVMC:
 		return 1;
-	case FOURCC_UYVY:
-	case FOURCC_YUY2:
 	default:
 		return 0;
 	}
 }
 
 bool
-sna_video_clip_helper(ScrnInfoPtr scrn,
-		      struct sna_video *adaptor_priv,
+sna_video_clip_helper(struct sna_video *video,
 		      struct sna_video_frame *frame,
-		      xf86CrtcPtr * crtc_ret,
+		      xf86CrtcPtr *crtc_ret,
 		      BoxPtr dst,
 		      short src_x, short src_y,
 		      short drw_x, short drw_y,
